
/* Sanfte Servos (Massen mit Modellbauservos ruckfrei bewegen)

Autor:	Frank Simon, post@ioproz.de (Danke an Malte Marwedel fr den Code 
		zur PWM, danke an Peter Fleury fr die uart-Routinen.)
Datum:	17.6.2005
Lizenz:	Es handelt sich bei diesem Code um ein Geschenk an die Menschheit.
		Keine Gewhr, ich hafte fr gar nichts!


Beschreibung:
Sanfte Servos: Bis zu 6 Achsen bewegen eine Masse sanft von einer Position zur 
nchsten. Nach Vorgabe des Zielpunktes und der Fahrzeit werden alle sechs Achsen 
mit konstanter Kraft beschleunigt. Ab der Hlfte des Weges wird mit derselben Kraft 
abgebremst. Die ntigen Berechnungen werden im Controller (AVR Mega8) ausgefhrt.
Hardware: Die Servos liegen an Port B0 bis B5, die Bewegungsparameter werden ber
RS232 bertragen. 
Das Protokoll ist simpel: Vom PC aus schreibt man immer zwei Bytes. Das erste ist
eine Adresse (nmlich der Index i fr das Byte servo_register[i] ), das zweite ist der
Wert, der gesetzt werden soll. Beispiel fr eine Bewegung zur Position 30,60,90,3,8,10
in 1 Sekunde:

Byte	Symbol				Bemerkung

15		REG_NEXTPOS + 1		Servo 1 an B0
30							Position 30 (Einheit: 90/256 = ca. 0,35)
16		REG_NEXTPOS + 2		Servo 2 an B1
60							Position 60
17		(usw.)				(usw.)
90
18
3
19
8
20		REG_NEXTPOS + 6		Servo 6 an B5
10
12		REG_TT				Zeitspanne, obere 8 Bit			
15
13		REG_TT+1			Zeitspanne, untere 8 Bit
66							Ergibt 1s, denn (15*256+66)*256us = 1s
14		REG_NEXTPOS			Eine 1 an dieser Stelle startet den
1							neuen Bewegungsablauf.

Der uC meldet jede Zwischenposition zurck mit 7 Bytes, die er sendet:
0 (zur Synchronisation), dann die 6 aktuellen Positionen.
Wenn ein Bewegungsablauf beendet wurde, sendet er statt einer Null
zwei Nullen. Erst dann darf REG_NEXTPOS wieder mit 1 beschrieben werden,
um den nchsten Bewegungsvorgang zu starten.

*/


#include <stdlib.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <math.h>

#include "uart.h"

#define UART_BAUD_RATE      38400   

#define SERVO_VERSION   0x03
#define SERVO_MAX       6

// Das Register ist der RAM-Bereich, der ber die RS232-Schnittstelle aktualisiert werden kann.
#define REG_VERSION     0x00						// Adresse 0: Versionsnummer
													// Adr.1 bis SERVO_MAX: aktueller Positionsvektor
#define REG_DELTA       SERVO_MAX + 1				// dann die Parameter
#define REG_OFFSETH     SERVO_MAX + 2
#define REG_OFFSETL     SERVO_MAX + 3
#define REG_REPEAT      SERVO_MAX + 4
#define REG_PORTC		SERVO_MAX + 5
#define REG_TT			SERVO_MAX + 6				// Zeitdauer der Bewegung, uint16_t in Vielfachen von 256us
#define REG_NEXTPOS		SERVO_MAX + 8				// =1 startet Neuberechnung, =2 zeigt an, dass sie fertig ist
													// Adr. REG_NEXTPOS+1 bis REG_NEXTPOS+SERVO_MAX: Positionsvektor neu
#define REG_ENDE		REG_NEXTPOS + SERVO_MAX		// Der letzte Eintrag im Register

#define DEFAULT_DELTA   (F_CPU / 256000ul)
#define DEFAULT_OFFSET  (F_CPU / 1000ul)
#define CYCLETIME		256							// Zykluszeit in Vielfachen von 128us (8MHZ/1024)
#define DEFAULT_REPEAT  256-CYCLETIME

#define SERVO_REG_MAX   REG_ENDE + 1

volatile uint8_t servo_register[SERVO_REG_MAX];
volatile uint8_t servo_rollover = 0;

#define Status_Stop				0
#define Status_Beschleunigen	0x01
#define Status_Abbremsen		0x02

struct serv {					
	uint8_t Status;		// Jedes Servo hat seinen individuellen Zustand (s.o.)
    uint8_t x1;			// Startposition der Bewegung
    uint8_t x2;			// Zielposition der Bewegung
	uint8_t hub;		// Abs(x2-x1)
	uint8_t x;			// Position des Servos, whrend der Fahrt also zwischen Start- und Zielwert.
	uint8_t index;		// zur Berechnung der wurzeligen Zeitabstnde (Beschleungigung, Abbremsung)
	uint16_t dtmax;		// Legt die Startgeschwindigkeit dieses Servos fest. Zeiteinheit: 256us
	uint32_t dt;		// Zeitdifferenz bis zur nchsten Pulsbreitennderung in us
	uint32_t t;			// Zeitpunkt fr die nchste nderung der Pulsdauer in us
  };
  
volatile struct serv servo[SERVO_MAX+1];

volatile uint32_t time = 0;		// Aktuelle Zeit in Vielfachen von 256us
volatile uint32_t time_sync;		// Startzeitpunkt des aktuell laufenden Zyklusses in us
volatile double dsqr[128];		// RAM fr Tabelle der wurzeligen Abstnde
volatile uint32_t count = 0;		// vllig unntig
volatile uint8_t NextAblaufFlag = 0;	// Zeigt an, dass am Ende des vorigen Frames alle
									// Servos still standen.


static void reset(void)
{
    uint8_t i;

    for (i = 0; i < SERVO_REG_MAX; i++)
		servo_register[i] = 0;	
 	servo_register[1] = 0;	
 	servo_register[2] = 0;	
 	servo_register[3] = 0;	
 	servo_register[4] = 0;	
 	servo_register[5] = 0;	
 	servo_register[6] = 0;	
    servo_register[REG_VERSION] = SERVO_VERSION;
    servo_register[REG_DELTA] = DEFAULT_DELTA;
    servo_register[REG_OFFSETH] = DEFAULT_OFFSET >> 8;
    servo_register[REG_OFFSETL] = DEFAULT_OFFSET & 0xFF;
    servo_register[REG_REPEAT] = DEFAULT_REPEAT;
	servo_register[REG_PORTC] = 0xFF;
	servo_register[REG_TT] = 15;
	servo_register[REG_TT+1] = 1;
	servo_register[REG_NEXTPOS+1] = 30;		// Hier steht die Startposition der Servos
	servo_register[REG_NEXTPOS+2] = 60;
	servo_register[REG_NEXTPOS+3] = 90;
	servo_register[REG_NEXTPOS+4] = 120;
	servo_register[REG_NEXTPOS+5] = 150;
	servo_register[REG_NEXTPOS+6] = 180;
	servo_register[REG_NEXTPOS] = 1;		// Hier steht, dass noch gerechnet werden muss
}



SIGNAL(SIG_OVERFLOW0)				// Timer null: Sytemzeit
{	time++;							// Es sind mal wieder 256us um.
}

SIGNAL(SIG_OVERFLOW1)				// Dies Ergeignis tritt ein, wenn die Zeit fr einen Servopuls 
{									// abgelaufen ist. Er wird beendet und gegebenenfalls ein neuer
    static uint8_t i = 0;			// begonnen.
    uint16_t servo_delay;
 
    switch (i) {									// Den laufenden Servopuls beenden
		case 0: break;									// <-- Hier war noch keiner an.
		case 1: PORTB = PINB &!(1<<0); break;		// Servo an Port B.0
		case 2: PORTB = PINB &!(1<<1); break;		// usw.
		case 3: PORTB = PINB &!(1<<2); break;
		case 4: PORTB = PINB &!(1<<3); break;
		case 5: PORTB = PINB &!(1<<4); break;
		case 6: PORTB = PINB &!(1<<5); break;
	}

    for (i++; i <= SERVO_MAX; i++) {	    // Nchsten Servo bestimmen
		if (servo_register[i] != 0)
			goto next;
    }
    if (servo_rollover) {				// Die Zeit war zu kurz, die restlichen Servos werden 
	i = 1;								// bersprungen.
	servo_rollover = 0;
	goto next;
    }

    /* an dieser stelle wurde kein aktiver servo gefunden */
    TCCR1B = 0; 						// Timer 1 stoppen, warten auf den Timer2-berlauf.
    i = 0;
    return;

next:
    servo_delay = 0;
    servo_delay -=
	(servo_register[REG_OFFSETH] << 8) + servo_register[REG_OFFSETL];
    servo_delay -=
	(uint16_t) servo_register[i] *
	(uint16_t) servo_register[REG_DELTA];
    TCNT1 = servo_delay;

    switch (i) {						// Den nchsten Servopuls starten
		case 1: PORTB = PINB | (1<<0); break;
		case 2: PORTB = PINB | (1<<1); break;
		case 3: PORTB = PINB | (1<<2); break;
		case 4: PORTB = PINB | (1<<3); break;
		case 5: PORTB = PINB | (1<<4); break;
		case 6: PORTB = PINB | (1<<5); break;
    }
}

SIGNAL(SIG_OVERFLOW2)
{	uint8_t i;
	uint8_t Status_Summe = 0;			// Zeigt an, ob alle Servos still stehen.

	for (i=1; i<=SERVO_MAX; i++)
		Status_Summe += servo[i].Status;
	if ( (Status_Summe == 0) )
	{	uart_putc(0);					// Statusmeldung absenden, falls alle Servos stehen
		if (servo_register[REG_NEXTPOS] == 1)
			NextAblaufFlag = 1;			// NextAblaufFlag setzen, falls alle Servos stehen und
	}									// der PC die neue Position gesetzt hat

	time_sync = (time<<8) + TCNT0;

	for ( i=1; i<=SERVO_MAX; i++ )	// Die neu berechneten Zwischenwerte aktivieren
		servo_register[i] = servo[i].x;
    if ((TCCR1B & 0x07) != 0) { 		// Falls Timer 1 noch luft...
	servo_rollover = 1;					// sind noch nicht alle Servos durch.
    }
    TCCR1B = 1; 						// Jedenfalls soll er ab jetzt (ohne Prescaler) wieder laufen.
    TCNT2 = servo_register[REG_REPEAT];// Zykluszeit fr den Gesamtablauf laden.
	TCNT1H = 255;						// und Impulserzeugung sehr bald starten.
	TCNT1L = 0;
}


void AblaufBerechnung(void)
{	// Wenn eine neue Position angefahren werden soll, steht der neue Wert frs Servo y in 
	// servo_register[REG_NEXTPOS+y]. Falls servo_register[REG_NEXTPOS] == 1, berechne neuen Bewegungsablauf!
	uint8_t i;
	
	if ( NextAblaufFlag == 1 )
	{	NextAblaufFlag = 0;
		for (i=1; i<=SERVO_MAX; i++) 	// Fr jedes Servo:
		{	// Falls eine neue Bewegung ntig ist, ...
			if ((servo_register[REG_NEXTPOS+i] != servo_register[i]) && (servo_register[REG_NEXTPOS+i] != 0))
			{	// ... dann den Servozustand neu berechnen:
				servo[i].x1 = servo_register[i];				// Start
				servo[i].x2 = servo_register[REG_NEXTPOS+i];	// Ende
				if ( servo[i].x2 > servo[i].x1 )				// Abstand
					servo[i].hub = servo[i].x2 - servo[i].x1;
				else
					servo[i].hub = servo[i].x1 - servo[i].x2;
				servo[i].x = servo[i].x1;						// aktueller Ort
				// In der folgenden Zeile steckt ganz furchtbar viel Physik...
				servo[i].dtmax = ((servo_register[REG_TT]<<8) + servo_register[REG_TT+1])/(sqrt(servo[i].hub<<1));
				servo[i].index = 0;
				servo[i].Status = Status_Beschleunigen;
				servo[i].t = time_sync;						// Zeitstempel in Mikrosekunden
			}
		}
		servo_register[REG_NEXTPOS] ++;	// =2, d.h. Bewegungsablauf ist berechnet. Start!
//		uart_putc(0);						// Statusmeldung absenden
	}
}



void Testdaten(void)
{
	if (servo_register[REG_NEXTPOS] == 0) {
		servo_register[REG_NEXTPOS+1] = 1;
		servo_register[REG_NEXTPOS+2] = 255;
		servo_register[REG_NEXTPOS+3] = 255;
		servo_register[REG_NEXTPOS+4] = 25;
		servo_register[REG_NEXTPOS+5] = 251;
		servo_register[REG_NEXTPOS+6] = 128;
		servo_register[REG_TT] = 15;				// Dauer der Bewegung: 1000ms = 3906 * Tick
		servo_register[REG_TT+1] = 66;
//			servo_register[REG_TT] = 7;						// Dauer der Bewegung: 500ms = 1953 * Tick
//			servo_register[REG_TT+1] = 161;
//			servo_register[REG_TT] = 1;						// Dauer der Bewegung: 100ms = 391 * Tick
//			servo_register[REG_TT+1] = 135;
		servo_register[REG_NEXTPOS] = 1;
	}
}




void RS232Auswertung(void)		// Ankommende Daten werden in das servo_register geschrieben
{	uint16_t c;
	uint8_t cc;
	static uint8_t reg = SERVO_REG_MAX;
	/* Get received character from ringbuffer
	 * uart_getc() returns in the lower byte the received character and 
	 * in the higher byte (bitmask) the last receive error
	 * UART_NO_DATA is returned when no data is available. */
	c = uart_getc();
	if ( c & UART_NO_DATA )
	{  // no data available from UART 
	}
	else
	{	/* new data available from UART
		 * check for Frame or Overrun error */
		if ( c & UART_FRAME_ERROR )
		{	/* Framing Error detected, i.e no stop bit detected */
			uart_puts_P("UART Frame Error: ");
		}
		if ( c & UART_OVERRUN_ERROR )
		{   /* Overrun, a character already present in the UART UDR register was 
			 * not read by the interrupt handler before the next character arrived,
			 * one or more received characters have been dropped */
			uart_puts_P("UART Overrun Error: ");
		}
		if ( c & UART_BUFFER_OVERFLOW )
		{   /* We are not reading the receive buffer fast enough,
			 * one or more received character have been dropped */
			uart_puts_P("Buffer overflow error: ");
		}
		cc = (unsigned char)c;			// received character in lower Byte
//		uart_putc( cc );					// send received character back
		if (reg < SERVO_REG_MAX) 			//Reg gibt an, wo das empfangene Byte hingeschrieben werden soll:
		{	servo_register[reg] = cc;		//Reg < SERVO_REG_MAX : Neuer Eintrag fr Nr. reg
			reg = SERVO_REG_MAX;			//und vermerken, dass nun wieder auf eine Adresse gewartet wird.
		}
		else
		{	reg = cc;						//Reg >= SERVO_REG_MAX : Neue Schreibaddr. in Reg merken.
		}	
	}
}


void SchrittBerechnung(void)
{	uint8_t i;

	for ( i=1; i<= SERVO_MAX; i++ ) {		// Fr jedes Servo:
		// Falls es Zeit ist, das Servo zu verschieben:
		while ( (servo[i].t <= time_sync) && (servo[i].Status != Status_Stop) ) 
		{					
			if ( servo[i].x < servo[i].x2 )	
				servo[i].x++;					// dann neue Position, rechter oder linker Nachbar
			if ( servo[i].x > servo[i].x2 )	
				servo[i].x--;
			
			if ( servo[i].Status == Status_Beschleunigen)				// Whrend der Beschleunigungsphase:
			{	if ( ((servo[i].index) << 1) >= servo[i].hub ) {		//Hlfte des Weges erreicht?
					servo[i].Status = Status_Abbremsen;				// Dann ab jetzt abbremsen.
					(servo[i].index)--;
					if ( servo[i].hub & 1 )
						(servo[i].index)--;		// Sonderfall ungerade Schrittdifferenz beachten.
				}
			}

			if ( servo[i].x == servo[i].x2)		// Ende erreicht, 
			{	servo[i].Status = Status_Stop;		// dann in Status vermerken.
				servo[i].index =0;
			}
			else
			{	// Nchsten Zeitpunkt berechnen, an dem das Servo eine Position weiter sein soll:
				// In der folgenden Zeile steckt nochmal ganz furchtbar viel Physik...			
				servo[i].dt = (((long) servo[i].dtmax) << 8) * dsqr[servo[i].index];
				servo[i].t += servo[i].dt;
	
				// Nchsten Tabelleneintrag vorbereiten fr die Ermittlung des nchsten Zeitintervalls:
				if ( servo[i].Status == Status_Beschleunigen )
					(servo[i].index)++;
				if ( servo[i].Status == Status_Abbremsen )
					if ( servo[i].index != 0)
						(servo[i].index)--;
			}
		if ( servo[i].x == servo[i].x2)			// Ende erreicht, 
			{	servo[i].Status = Status_Stop;		// dann in Status vermerken.
				servo[i].index =0;
			}
		}
	}
}


int main(void)
{
    uint8_t i;
	uint32_t time_sync_old = 0;

	for ( i=0; i<128; i++)					// Die Tabelle der Wurzelabstnde ist fr Bewegungen mit
		dsqr[i] = sqrt(i+1) - sqrt(i);		// konstanter Beschleunigung praktisch.

    reset();					// Registerbereich initialisieren

    DDRB = 0b00111111;			//PB6 und PB7 nicht verwenden wegen Quarz
    PORTB = 0x00;
    DDRC = 0xFF;				// Port C als Universalausgnge

    uart_init( UART_BAUD_SELECT(UART_BAUD_RATE,F_CPU) ); 	// USART vorbereiten

	TCCR0 =2;												// Timer0 mit Vorteiler 8 zhlt Mikrosekunden
    TCCR1B = 0; 											// Timer1 steht anfangs still
    TCCR2 = 7; 												// Timer2 mit Vorteiler 1024
    TCNT2 = 0;//250;	
    TIMSK |= _BV(TOIE2)| _BV(TOIE1)| _BV(TOIE0);			// berlauf-Interrupts freigeben

	Testdaten();
	NextAblaufFlag = 1;
	AblaufBerechnung();

 	sei();
//    uart_puts("Bereit.");

	// set_sleep_mode(SLEEP_MODE_IDLE);

	for (;;)
		{	
			count = 0;							// Beginn eines neuen Frames
	
			SchrittBerechnung();				// Schrittdaten fr nchsten Frame berechnen
	
			AblaufBerechnung();					// ggf. neuen Bewegungsablauf vorbereiten
	
			uart_putc(0);						// Statusmeldung absenden
			for (i=1; i<=SERVO_MAX; i++ ) 	// Aktuelle Position
				uart_putc( servo_register[i] );
			
			if (servo_register[REG_VERSION] != SERVO_VERSION) reset();	// ein wenig Kleinkram...
			// PORTC = servo_register[REG_PORTC];
	
			// So, ab jetzt wird auf das Ende des Frames gewartet.
			PORTC = 0; 							// Led an zeigt die verbleibende Rechenkapazitt an
			while (time_sync == time_sync_old) // Ab jetzt Langeweile fr diesen Frame
			{	count++;						// count zhlt nur so vor sich hin.
				RS232Auswertung();				// ggf. servo_register aktualisieren ...
			}
			PORTC = 255;  						// Led aus

			time_sync_old = time_sync;			// Hier landen wir erst, wenn der Frame vorbei ist.
		}
}
